// Disk Imager 1.00.98.2
// By Steven Lawrance

#include <windows.h>
#include <commctrl.h>
#include <winioctl.h>
#include "resource.h"


// DiskOps Structure
typedef struct _DISKOPS {
  LPTSTR lpszImagePath;
  LPTSTR lpszDiskPath;
  int method;
  HWND hDlg;
  BOOL clear;
} DISKOPS, *LPDISKOPS;


// ChangeStatus Function
void ChangeStatus(HWND hDlg, LPCTSTR status, DWORD num1, DWORD num2) {
  LPSTR str = (LPSTR) GlobalAlloc(GPTR, lstrlen(status) + 21);
  wsprintf(str, status, num1, num2);
  SetDlgItemText(hDlg, IDC_STATUS, str);
  GlobalFree(GlobalHandle(str));
}

// DisplayError Function
void DisplayError(HWND hDlg, DWORD num, LPCTSTR str, LPCTSTR info, BOOL systemError) {
  // Allocate and assign the error message string
  LPTSTR msg;
  if (systemError) {
    // This is a system error - get the last error string
    LPTSTR errorMsg;
    FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM, NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &errorMsg, 1, NULL);
    msg = (LPTSTR) GlobalAlloc(GPTR, lstrlen(str) + lstrlen(info) + lstrlen(errorMsg) + 29);
    wsprintf(msg, "%s (%lu.%lu).\n\n%s\n%s", str, num, GetLastError(), errorMsg, info);
    LocalFree(LocalHandle(errorMsg));
  } else {
    // Not a system error - just allocate what we need and leave :-)
    msg = (LPTSTR) GlobalAlloc(GPTR, lstrlen(str) + lstrlen(info) + 17);
    wsprintf(msg, "%s (%lu).\n\n%s", str, num, info);
  }

  // Display the message
  MessageBox(hDlg, msg, "Disk Imager", MB_OK | MB_ICONEXCLAMATION);

  // Free the memory
  GlobalFree(GlobalHandle(msg));
}

// Changes the enabled state of the dialog items
void ChangeEnabledState(HWND hDlg, BOOL on) {
  DWORD i;
  if (on) {
    // Copy to Disk is selected
    if (IsDlgButtonChecked(hDlg, IDC_COPYTODISK) == BST_CHECKED) {
      for (i = IDC_DISKDRIVE; i <= IDC_ERASEDISK; i++)
        EnableWindow(GetDlgItem(hDlg, i), true);
    }

    // Either Copy to Image File or Verify Disk is selected
    if ((IsDlgButtonChecked(hDlg, IDC_COPYTOFILE) == BST_CHECKED) ||
        (IsDlgButtonChecked(hDlg, IDC_VERIFYDISK) == BST_CHECKED)) {
      for (i = IDC_DISKDRIVE; i <= IDC_ERASEDISK; i++)
        if (i == IDC_CLEARSPACE)
          EnableWindow(GetDlgItem(hDlg, i), false);
        else
          EnableWindow(GetDlgItem(hDlg, i), true);
    }

    // Erase Disk is selected
    if (IsDlgButtonChecked(hDlg, IDC_ERASEDISK) == BST_CHECKED) {
      for (i = IDC_DISKDRIVE; i <= IDC_ERASEDISK; i++)
        if ((i == IDC_CLEARSPACE) || (i == IDC_IMAGEPATH) || (i == IDC_BROWSEFILES))
          EnableWindow(GetDlgItem(hDlg, i), false);
        else
          EnableWindow(GetDlgItem(hDlg, i), true);
    }

    // Show the Exit button
    EnableWindow(GetDlgItem(hDlg, IDCANCEL), true);
  } else {
    // Disable all dialog items
    for (i = IDC_DISKDRIVE; i <= IDC_ERASEDISK; i++)
      EnableWindow(GetDlgItem(hDlg, i), false);
    EnableWindow(GetDlgItem(hDlg, IDCANCEL), false);
  }
}


// Thread Handles
HANDLE hThread;
DWORD threadID;
BOOL stopDiskOps;


// DiskOps Function
int DiskOps(LPDISKOPS lpDiskOps) {
  HANDLE hImageFile = INVALID_HANDLE_VALUE;
  HANDLE hDiskFile = INVALID_HANDLE_VALUE;
  BOOL run = true;

  // Set the "Run" button text to "Stop"
  SetDlgItemText(lpDiskOps->hDlg, IDC_RUN, "&Stop");

  // Switch for method
  switch (lpDiskOps->method) {
    
  case IDC_COPYTODISK :
    // Ensure the user wants to write to the disk
    LPSTR warningStr;
    if (lpDiskOps->clear)
      warningStr = "WARNING!  This operation will erase all data on the disk and write the selected image file to the disk!\n\nAre you sure you want to copy the image to the disk?";
    else
      warningStr = "WARNING!  This operation will replace data on the disk with the selected image file, but does not affect data past the end of the image file (if any).\n\nAre you sure you want to copy the image to the disk?";
    if (MessageBox(lpDiskOps->hDlg, warningStr, "Disk Imager", MB_YESNO | MB_DEFBUTTON2 | MB_ICONQUESTION) == IDNO)
      run = false;
    else {
      // Load the files if okay
      hImageFile = CreateFile(lpDiskOps->lpszImagePath, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
      hDiskFile = CreateFile(lpDiskOps->lpszDiskPath, GENERIC_WRITE | GENERIC_READ, FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);
      
      // Check to see if the files opened
      if (hImageFile == INVALID_HANDLE_VALUE) {
        DisplayError(lpDiskOps->hDlg, 2, "Image file cannot be opened", "Make sure the image file exists, and that no program has the image file opened for exclusive access.", true);
        run = false;
      }
      if (hDiskFile == INVALID_HANDLE_VALUE) {
        DisplayError(lpDiskOps->hDlg, 3, "Could not open the target disk", "Be sure that a disk is in the drive, its write-protect window is closed, no files on the disk are opened, and that all explorer windows on the disk are closed.", true);
        run = false;
      }
    }

    if (run) {
      // Copy the image file to the disk
      DISK_GEOMETRY geometry;
      DWORD readBytes = 0, writtenBytes;
      
      // Get the disk geometry
      if (!DeviceIoControl(hDiskFile, IOCTL_DISK_GET_DRIVE_GEOMETRY, NULL, 0, &geometry, sizeof(geometry), &writtenBytes, NULL)) {
        DisplayError(lpDiskOps->hDlg, 50, "Could not get disk geometry", "Make sure the disk is in the disk drive and that you are not trying to write to a CD-R drive.", true);
      } else {
        // Obtain exclusive access to the disk
        if (!DeviceIoControl(hDiskFile, FSCTL_LOCK_VOLUME, NULL, 0, NULL, 0, &writtenBytes, NULL))
          DisplayError(lpDiskOps->hDlg, 4, "Could not obtain exclusive access to the target disk", "Be sure that no files on the disk are opened and that all explorer windows on the disk are closed.", true);
        else {
          DWORD bufSize = geometry.TracksPerCylinder * geometry.SectorsPerTrack * geometry.BytesPerSector;
          LPBYTE buf = (LPBYTE) VirtualAlloc(NULL, bufSize, MEM_COMMIT, PAGE_READWRITE);
          DWORD numCyln, size = 0, percentDone = 0, percentDisplayed = 1;
          DWORD fileSize = SetFilePointer(hImageFile, 0, NULL, FILE_END);
          SetFilePointer(hImageFile, 0, NULL, FILE_BEGIN);

          // Align fileSize to a cylinder boundry, then make fileSize express number of cylinders
          if (fileSize % bufSize != 0)
            fileSize = (fileSize / bufSize + 1) * bufSize;
          fileSize = fileSize / bufSize;
          
          // Copy data to the disk in cylinders
          ChangeStatus(lpDiskOps->hDlg, "Copying image file to disk", 0, 0);
          ChangeEnabledState(lpDiskOps->hDlg, false);
          numCyln = geometry.Cylinders.LowPart;
          do {
            // Show the percentage, if changed
            percentDone = ((geometry.Cylinders.LowPart - numCyln) * 100) / fileSize;
            if (percentDone != percentDisplayed) {
              percentDisplayed = percentDone;
              ChangeStatus(lpDiskOps->hDlg, "Copying image file to disk - (%lu%%)", percentDone, 0);
              SendDlgItemMessage(lpDiskOps->hDlg, IDC_PROGRESS, PBM_SETPOS, percentDone, 0);
            }
            
            // Read the file
            if (!ReadFile(hImageFile, buf, bufSize, &readBytes, NULL)) {
              DisplayError(lpDiskOps->hDlg, 5, "Could not read file", "Make sure the disk with the source image file is okay.", true);
              stopDiskOps = true;
            }
            
            if ((bufSize != readBytes) && !stopDiskOps) {
              if (lpDiskOps->clear) {
                // If the file does not have enough for the cylinder, then fill in the rest
                // with CHR(246)
                FillMemory((LPBYTE) (buf + readBytes + 1), bufSize - readBytes, 246);
              } else {
                // Read the current disk cylinder
                if (!ReadFile(hDiskFile, buf, bufSize, &writtenBytes, NULL)) {
                  DisplayError(lpDiskOps->hDlg, 6, "Could not read from the disk while attempting a seamless image boundry", "The disk may have existing damage on it.  Try another disk.  At this time, Disk Imager does not support imperfect disks (sorry).", true);
                  stopDiskOps = true;
                } else {
                  // Go back in the file and read again
                  SetFilePointer(hDiskFile, -((LONG) bufSize), NULL, FILE_CURRENT);
                  SetFilePointer(hImageFile, -((LONG) readBytes), NULL, FILE_CURRENT);
                  if (!ReadFile(hImageFile, buf, bufSize, &readBytes, NULL)) {
                    DisplayError(lpDiskOps->hDlg, 7, "Could not read file while attempting a seamless image boundry", "Make sure the disk with the source image file supports file seeking (all drives should, but some TCP/IP connections might not).", true);
                    stopDiskOps = true;
                  }
                }
              }
            }
            
            // Write to the disk, if we read anything
            if ((readBytes > 0) && !stopDiskOps) {
              if (WriteFile(hDiskFile, buf, bufSize, &writtenBytes, NULL))
                size += readBytes;
              else {
                DisplayError(lpDiskOps->hDlg, 8, "Could not write to the disk", "Be sure its write-protect window is closed.  If it is closed and you got this message, the disk may have existing damage on it.  Try another disk.  At this time, Disk Imager does not support imperfect disks (sorry).", true);
                stopDiskOps = true;
              }
              numCyln--;
            }
          } while ((readBytes == bufSize) && (numCyln > 0) && !stopDiskOps);
          
          // Fill the rest of the disk with CHR(246)
          percentDone = 0;
          fileSize = numCyln;
          if ((lpDiskOps->clear) && !stopDiskOps) {
            FillMemory(buf, bufSize, 246);
            while ((numCyln > 0) && !stopDiskOps) {
              // Show the percentage, if changed
              percentDone = ((fileSize - numCyln) * 100) / fileSize;
              if (percentDone != percentDisplayed) {
                percentDisplayed = percentDone;
                ChangeStatus(lpDiskOps->hDlg, "Writing empty space to disk - (%lu%%)", percentDone, 0);
                SendDlgItemMessage(lpDiskOps->hDlg, IDC_PROGRESS, PBM_SETPOS, percentDone, 0);
              }
              
              // Write the "empty" space to the disk          
              if (!WriteFile(hDiskFile, buf, bufSize, &writtenBytes, NULL)) {
                DisplayError(lpDiskOps->hDlg, 9, "Could not write empty space to the disk", "Be sure its write-protect window is closed.  If it is closed and you got this message, the disk may have existing damage on it.  Try another disk.  At this time, Disk Imager does not support imperfect disks (sorry).", true);
                stopDiskOps = true;
              }
              numCyln--;
            }
          }
          
          // Free the buffer
          VirtualFree((LPVOID) buf, bufSize, MEM_DECOMMIT);
          
          // Unlock the drive
          writtenBytes = 0;
          DeviceIoControl(hDiskFile, FSCTL_UNLOCK_VOLUME, NULL, 0, NULL, 0, &writtenBytes, NULL);
          
          // Write the result (test to see if the file has more data in it too)
          if (SetFilePointer(hImageFile, 0, NULL, FILE_CURRENT) == SetFilePointer(hImageFile, 0, NULL, FILE_END))
            ChangeStatus(lpDiskOps->hDlg, "Copied %lu byte(s) to the disk", size, 0);
          else
            ChangeStatus(lpDiskOps->hDlg, "First %lu byte(s) of image file written to the disk", size, 0);
          ChangeEnabledState(lpDiskOps->hDlg, true);
          SendDlgItemMessage(lpDiskOps->hDlg, IDC_PROGRESS, PBM_SETPOS, 0, 0);
        }
      }
    }
    break;
    
  case IDC_COPYTOFILE :
    // See if the file exists; if so, ask the user for permission to overwrite
    hImageFile = CreateFile(lpDiskOps->lpszImagePath, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
    if (hImageFile != INVALID_HANDLE_VALUE) {
      if (MessageBox(lpDiskOps->hDlg, "File already exists!  Do you want to overwrite the file with the contents of the disk?", "Disk Imager", MB_YESNO | MB_DEFBUTTON2 | MB_ICONQUESTION) == IDNO)
        run = false;
      CloseHandle(hImageFile);
    }

    if (run) {
      // Load the files if okay
      hImageFile = CreateFile(lpDiskOps->lpszImagePath, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL);
      hDiskFile = CreateFile(lpDiskOps->lpszDiskPath, GENERIC_READ, FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);
      
      // Check to see if the files opened
      if (hImageFile == INVALID_HANDLE_VALUE) {
        DisplayError(lpDiskOps->hDlg, 10, "Image file cannot be created", "Make sure the image file exists, and that no program has the image file opened for exclusive access.", true);
        run = false;
      }
      if (hDiskFile == INVALID_HANDLE_VALUE) {
        DisplayError(lpDiskOps->hDlg, 11, "Could not open the source disk", "Be sure that a disk is in the drive, no files on the disk are opened, and that all explorer windows on the disk are closed.", true);
        run = false;
      }
    }
    
    if (run) {
      // Copy the disk to the image file
      DISK_GEOMETRY geometry;
      DWORD readBytes = 0, writtenBytes;

      // Get the disk geometry
      if (!DeviceIoControl(hDiskFile, IOCTL_DISK_GET_DRIVE_GEOMETRY, NULL, 0, &geometry, sizeof(geometry), &writtenBytes, NULL)) {
        DisplayError(lpDiskOps->hDlg, 51, "Could not get disk geometry", "Make sure the disk is in the disk drive and that you are not trying to read a CD-ROM drive.", true);
      } else {
        // Obtain exclusive access to the disk
        if (!DeviceIoControl(hDiskFile, FSCTL_LOCK_VOLUME, NULL, 0, NULL, 0, &writtenBytes, NULL))
          DisplayError(lpDiskOps->hDlg, 12, "Could not obtain exclusive access to the source disk", "Be sure that no files on the disk are opened and that all explorer windows on the disk are closed.", true);
        else {
          DWORD bufSize = geometry.TracksPerCylinder * geometry.SectorsPerTrack * geometry.BytesPerSector;
          LPBYTE buf = (LPBYTE) VirtualAlloc(NULL, bufSize, MEM_COMMIT, PAGE_READWRITE);
          DWORD numCyln, size = 0, percentDone = 0, percentDisplayed = 1;
          
          // Copy data to the disk in cylinders
          ChangeStatus(lpDiskOps->hDlg, "Copying disk to image file", 0, 0);
          ChangeEnabledState(lpDiskOps->hDlg, false);
          numCyln = geometry.Cylinders.LowPart;
          do {
            // Show the percentage, if changed
            percentDone = ((geometry.Cylinders.LowPart - numCyln) * 100) / geometry.Cylinders.LowPart;
            if (percentDone != percentDisplayed) {
              percentDisplayed = percentDone;
              ChangeStatus(lpDiskOps->hDlg, "Copying disk to image file - (%lu%%)", percentDone, 0);
              SendDlgItemMessage(lpDiskOps->hDlg, IDC_PROGRESS, PBM_SETPOS, percentDone, 0);
            }
            
            // Read the disk
            if (!ReadFile(hDiskFile, buf, bufSize, &readBytes, NULL)) {
              DisplayError(lpDiskOps->hDlg, 13, "Could not read the disk", "The disk may have existing damage on it.  Try another disk.  At this time, Disk Imager does not support imperfect disks (sorry).", true);
              stopDiskOps = true;
            }
            
            // Write to the file, if we read anything
            if ((readBytes > 0) && !stopDiskOps) {
              if (WriteFile(hImageFile, buf, readBytes, &writtenBytes, NULL))
                size += readBytes;
              else {
                DisplayError(lpDiskOps->hDlg, 14, "Could not write to the disk", "Make sure you have enough space on the drive with the target image file.  If you have enough space, check your disk for errors.", true);
                stopDiskOps = true;
              }
              numCyln--;
            }
          } while ((readBytes > 0) && (numCyln > 0) && !stopDiskOps);
          
          // Free the buffer
          VirtualFree((LPVOID) buf, bufSize, MEM_DECOMMIT);
          
          // Unlock the drive
          writtenBytes = 0;
          DeviceIoControl(hDiskFile, FSCTL_UNLOCK_VOLUME, NULL, 0, NULL, 0, &writtenBytes, NULL);
          
          // Write the result
          ChangeStatus(lpDiskOps->hDlg, "Copied %lu byte(s) to the image file", size, 0);
          ChangeEnabledState(lpDiskOps->hDlg, true);
          SendDlgItemMessage(lpDiskOps->hDlg, IDC_PROGRESS, PBM_SETPOS, 0, 0);
        }
      }
    }
    break;

  case IDC_ERASEDISK :
    // Ensure the user wants to erase the disk
    if (MessageBox(lpDiskOps->hDlg, "WARNING!  This operation will erase all data on the disk!\n\nAre you sure you want to do this?", "Disk Imager", MB_YESNO | MB_DEFBUTTON2 | MB_ICONQUESTION) == IDNO)
      run = false;
    else {
      // Load the file if okay
      hDiskFile = CreateFile(lpDiskOps->lpszDiskPath, GENERIC_WRITE, FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);
      
      // Check to see if the files opened
      if (hDiskFile == INVALID_HANDLE_VALUE) {
        DisplayError(lpDiskOps->hDlg, 20, "Could not open the target disk", "Be sure that a disk is in the drive, it's write-protect window is closed, no files on the disk are opened, and that all explorer windows on the disk are closed.", true);
        run = false;
      }
    }

    if (run) {
      // Erase the disk
      DISK_GEOMETRY geometry;
      DWORD writtenBytes;

      // Get the disk geometry
      if (!DeviceIoControl(hDiskFile, IOCTL_DISK_GET_DRIVE_GEOMETRY, NULL, 0, &geometry, sizeof(geometry), &writtenBytes, NULL)) {
        DisplayError(lpDiskOps->hDlg, 52, "Could not get disk geometry", "Make sure the disk is in the disk drive and that you are not trying to write to a CD-R drive.", true);
      } else {
        // Obtain exclusive access to the disk
        if (!DeviceIoControl(hDiskFile, FSCTL_LOCK_VOLUME, NULL, 0, NULL, 0, &writtenBytes, NULL))
          DisplayError(lpDiskOps->hDlg, 21, "Could not obtain exclusive access to the disk", "Be sure that no files on the disk are opened and that all explorer windows on the disk are closed.", true);
        else {
          DWORD bufSize = geometry.TracksPerCylinder * geometry.SectorsPerTrack * geometry.BytesPerSector;
          LPBYTE buf = (LPBYTE) VirtualAlloc(NULL, bufSize, MEM_COMMIT, PAGE_READWRITE);
          DWORD numCyln, size = 0, percentDone = 0, percentDisplayed = 1;
          
          // Fill the buffer with CHR(246)
          ChangeStatus(lpDiskOps->hDlg, "Erasing the disk", 0, 0);
          ChangeEnabledState(lpDiskOps->hDlg, false);
          FillMemory(buf, bufSize, 246);
          
          // Copy data to the disk in cylinders
          numCyln = geometry.Cylinders.LowPart;
          do {
            // Show the percentage, if changed
            percentDone = ((geometry.Cylinders.LowPart - numCyln) * 100) / geometry.Cylinders.LowPart;
            if (percentDone != percentDisplayed) {
              percentDisplayed = percentDone;
              ChangeStatus(lpDiskOps->hDlg, "Erasing the disk - (%lu%%)", percentDone, 0);
              SendDlgItemMessage(lpDiskOps->hDlg, IDC_PROGRESS, PBM_SETPOS, percentDone, 0);
            }
            
            // Write the blank data to the disk
            if (WriteFile(hDiskFile, buf, bufSize, &writtenBytes, NULL))
              size += writtenBytes;
            else {
              DisplayError(lpDiskOps->hDlg, 22, "Could not write to the disk", "Be sure its write-protect window is closed.  If it is closed and you got this message, the disk may have existing damage on it.  Try another disk.  At this time, Disk Imager does not support imperfect disks (sorry).", true);
              stopDiskOps = true;
            }
            numCyln--;
          } while ((writtenBytes > 0) && (numCyln > 0) && !stopDiskOps);
          
          // Free the buffer
          VirtualFree((LPVOID) buf, bufSize, MEM_DECOMMIT);
          
          // Unlock the drive
          writtenBytes = 0;
          DeviceIoControl(hDiskFile, FSCTL_UNLOCK_VOLUME, NULL, 0, NULL, 0, &writtenBytes, NULL);
          
          // Write the result (test to see if the file has more data in it too)
          ChangeStatus(lpDiskOps->hDlg, "Erased %lu byte(s) on the disk", size, 0);
          ChangeEnabledState(lpDiskOps->hDlg, true);
          SendDlgItemMessage(lpDiskOps->hDlg, IDC_PROGRESS, PBM_SETPOS, 0, 0);
        }
      }
    }
    break;

  case IDC_VERIFYDISK :
    // Load the files if okay
    hImageFile = CreateFile(lpDiskOps->lpszImagePath, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
    hDiskFile = CreateFile(lpDiskOps->lpszDiskPath, GENERIC_READ, FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);
    
    // Check to see if the files opened
    if (hImageFile == INVALID_HANDLE_VALUE) {
      DisplayError(lpDiskOps->hDlg, 30, "Image file cannot be opened", "Make sure the image file exists, and that no program has the image file opened for exclusive access.", true);
      run = false;
    }
    if (hDiskFile == INVALID_HANDLE_VALUE) {
      DisplayError(lpDiskOps->hDlg, 31, "Could not open the source disk", "Be sure that a disk is in the drive, no files on the disk are opened, and that all explorer windows on the disk are closed.", true);
      run = false;
    }
    
    if (run) {
      // Verify the disk with the image file
      DISK_GEOMETRY geometry;
      DWORD readBytes = 0, writtenBytes;

      // Get the disk geometry
      if (!DeviceIoControl(hDiskFile, IOCTL_DISK_GET_DRIVE_GEOMETRY, NULL, 0, &geometry, sizeof(geometry), &writtenBytes, NULL)) {
        DisplayError(lpDiskOps->hDlg, 53, "Could not get disk geometry", "Make sure the disk is in the disk drive and that you are not trying to read a CD-ROM drive.", true);
      } else {
        // Obtain exclusive access to the disk
        if (!DeviceIoControl(hDiskFile, FSCTL_LOCK_VOLUME, NULL, 0, NULL, 0, &writtenBytes, NULL))
          DisplayError(lpDiskOps->hDlg, 32, "Could not obtain exclusive access to the source disk", "Be sure that no files on the disk are opened and that all explorer windows on the disk are closed.", true);
        else {
          DWORD bufSize = geometry.TracksPerCylinder * geometry.SectorsPerTrack * geometry.BytesPerSector;
          LPBYTE buf1 = (LPBYTE) VirtualAlloc(NULL, bufSize, MEM_COMMIT, PAGE_READWRITE);
          LPBYTE buf2 = (LPBYTE) VirtualAlloc(NULL, bufSize, MEM_COMMIT, PAGE_READWRITE);
          DWORD numCyln, percentDone = 0, percentDisplayed = 1;
          DWORD diffs = 0, i;
          
          // Verify disk data with the image file
          ChangeStatus(lpDiskOps->hDlg, "Verifying disk with image file", 0, 0);
          ChangeEnabledState(lpDiskOps->hDlg, false);
          numCyln = geometry.Cylinders.LowPart;
          do {
            // Show the percentage, if changed
            percentDone = ((geometry.Cylinders.LowPart - numCyln) * 100) / geometry.Cylinders.LowPart;
            if (percentDone != percentDisplayed) {
              percentDisplayed = percentDone;
              ChangeStatus(lpDiskOps->hDlg, "Verifying disk with image file - (%lu%%)", percentDone, 0);
              SendDlgItemMessage(lpDiskOps->hDlg, IDC_PROGRESS, PBM_SETPOS, percentDone, 0);
            }
            
            // Read the disk
            if (!ReadFile(hDiskFile, buf1, bufSize, &readBytes, NULL)) {
              DisplayError(lpDiskOps->hDlg, 33, "Could not read the disk", "The disk may have existing damage on it.  Try another disk.  At this time, Disk Imager does not support imperfect disks (sorry).", true);
              readBytes = 0;
            }
            
            // Read the file
            if (!ReadFile(hImageFile, buf2, bufSize, &readBytes, NULL)) {
              DisplayError(lpDiskOps->hDlg, 34, "Could not read file", "Make sure the disk with the source image file is okay.", true);
              readBytes = 0;
            }
            
            // Compare the buffers
            for (i = 0; i < readBytes; i++) {
              if (buf1[i] != buf2[i])
                diffs++;
            }
            
            // Next cylinder
            numCyln--;
          } while ((readBytes == bufSize) && (numCyln > 0) && !stopDiskOps);
          
          // Free the buffers
          VirtualFree((LPVOID) buf1, bufSize, MEM_DECOMMIT);
          VirtualFree((LPVOID) buf2, bufSize, MEM_DECOMMIT);
          
          // Unlock the drive
          writtenBytes = 0;
          DeviceIoControl(hDiskFile, FSCTL_UNLOCK_VOLUME, NULL, 0, NULL, 0, &writtenBytes, NULL);
          
          // Write the result
          DWORD pos = SetFilePointer(hImageFile, 0, NULL, FILE_CURRENT);
          if (pos == SetFilePointer(hImageFile, 0, NULL, FILE_END)) {
            // End of file reached
            if (pos < bufSize * geometry.Cylinders.LowPart) {
              // File is smaller than disk
              if (diffs)
                ChangeStatus(lpDiskOps->hDlg, "%lu byte(s) differ between image file and first %lu byte(s) of disk", diffs, pos);
              else
                ChangeStatus(lpDiskOps->hDlg, "First %lu byte(s) of disk match the image file", pos, 0);
            } else {
              // File is equal to disk (in size)
              if (diffs)
                ChangeStatus(lpDiskOps->hDlg, "%lu byte(s) differ between disk and image file", diffs, 0);
              else
                ChangeStatus(lpDiskOps->hDlg, "The image file and disk are identical", pos, 0);
            }
          } else {
            // End of file not reached
            if (pos < bufSize * geometry.Cylinders.LowPart) {
              // Operation had to have been canceled
              ChangeStatus(lpDiskOps->hDlg, "Operation canceled", 0, 0);
            } else {
              // File is larger than disk
              if (diffs)
                ChangeStatus(lpDiskOps->hDlg, "%lu byte(s) differ between disk and first %lu byte(s) of image file", diffs, pos);
              else
                ChangeStatus(lpDiskOps->hDlg, "First %lu byte(s) of the image file match disk", pos, 0);
            }
          }
          ChangeEnabledState(lpDiskOps->hDlg, true);
          SendDlgItemMessage(lpDiskOps->hDlg, IDC_PROGRESS, PBM_SETPOS, 0, 0);
        }
      }
    }
    break;

  }

  // Set the "Stop" button text to "Run"
  SetDlgItemText(lpDiskOps->hDlg, IDC_RUN, "&Run");

  // Close the handles, if open
  if (hDiskFile != INVALID_HANDLE_VALUE)
    CloseHandle(hDiskFile);
  if (hImageFile != INVALID_HANDLE_VALUE)
    CloseHandle(hImageFile);

  // Free the memory associated with the structure
  GlobalFree(GlobalHandle(lpDiskOps->lpszImagePath));
  GlobalFree(GlobalHandle(lpDiskOps->lpszDiskPath));
  GlobalFree(GlobalHandle(lpDiskOps));

  // Set the threadID to zero, indicating that we're done :-)!
  threadID = 0;
  return 0;
}


// DiskOps Dialog
BOOL CALLBACK DiskOpsDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  switch (uMsg) {
    
  case WM_INITDIALOG :
    // Initialize the dialog box elements
    DlgDirListComboBox(hDlg, "", IDC_DISKDRIVE, NULL, DDL_DRIVES);
    SendDlgItemMessage(hDlg, IDC_DISKDRIVE, CB_SETCURSEL, 0, 0);
    CheckDlgButton(hDlg, IDC_COPYTODISK, BST_CHECKED);
    CheckDlgButton(hDlg, IDC_CLEARSPACE, BST_CHECKED);
    SetDlgItemText(hDlg, IDC_IMAGEPATH, (LPTSTR) lParam);
    return true;
    
  case WM_COMMAND :
    switch (LOWORD(wParam)) {

    case IDC_COPYTODISK :
      // Enable the IDC_CLEARSPACE and image-related dialog items
      EnableWindow(GetDlgItem(hDlg, IDC_CLEARSPACE), true);
      EnableWindow(GetDlgItem(hDlg, IDC_IMAGEPATH), true);
      EnableWindow(GetDlgItem(hDlg, IDC_BROWSEFILES), true);
      break;

    case IDC_COPYTOFILE :
    case IDC_VERIFYDISK :
      // Disable the IDC_CLEARSPACE and enable the image-related dialog items
      EnableWindow(GetDlgItem(hDlg, IDC_CLEARSPACE), false);
      EnableWindow(GetDlgItem(hDlg, IDC_IMAGEPATH), true);
      EnableWindow(GetDlgItem(hDlg, IDC_BROWSEFILES), true);
      break;

    case IDC_ERASEDISK :
      // Disable the IDC_CLEARSPACE and image path-related dialog items
      EnableWindow(GetDlgItem(hDlg, IDC_CLEARSPACE), false);
      EnableWindow(GetDlgItem(hDlg, IDC_IMAGEPATH), false);
      EnableWindow(GetDlgItem(hDlg, IDC_BROWSEFILES), false);
      break;
      
    case IDC_RUN : {
      // If the thread is running, then confirm the stop
      if (threadID) {
        if (MessageBox(hDlg, "Are you sure you want to stop the current operation?", "Disk Imager", MB_YESNO | MB_DEFBUTTON2 | MB_ICONEXCLAMATION) == IDYES) {
          // Ask the thread to exit by setting the global exit flag
          stopDiskOps = true;
          SetDlgItemText(hDlg, IDC_STATUS, "Stopping current operation...");
        }
      } else {
        // Create a DiskOps structure
        LPDISKOPS lpDiskOps = (LPDISKOPS) GlobalAlloc(GPTR, sizeof(DISKOPS));
        
        // Get the image path
        lpDiskOps->lpszImagePath = (LPTSTR) GlobalAlloc(GPTR, sizeof(char) * SendDlgItemMessage(hDlg, IDC_IMAGEPATH, EM_LINELENGTH, 0, 0) + 1);
        GetDlgItemText(hDlg, IDC_IMAGEPATH, lpDiskOps->lpszImagePath, GlobalSize(GlobalHandle(lpDiskOps->lpszImagePath)));
        
        // Get the disk path
        lpDiskOps->lpszDiskPath = (LPTSTR) GlobalAlloc(GPTR, sizeof(char) * 7); // Example: "\\\\.\\a:", 7 chars w/ terminating NULL
        SendDlgItemMessage(hDlg, IDC_DISKDRIVE, CB_GETLBTEXT, SendDlgItemMessage(hDlg, IDC_DISKDRIVE, CB_GETCURSEL, 0, 0), (LPARAM) lpDiskOps->lpszDiskPath);
        lpDiskOps->lpszDiskPath[4] = lpDiskOps->lpszDiskPath[2];
        CopyMemory(lpDiskOps->lpszDiskPath, "\\\\.\\", 4);
        lpDiskOps->lpszDiskPath[5] = ':';
        lpDiskOps->lpszDiskPath[6] = 0;
        
        // Get the method from the radio buttons
        lpDiskOps->method = 0;
        if (IsDlgButtonChecked(hDlg, IDC_COPYTODISK) == BST_CHECKED)
          lpDiskOps->method = IDC_COPYTODISK;
        if (IsDlgButtonChecked(hDlg, IDC_COPYTOFILE) == BST_CHECKED)
          lpDiskOps->method = IDC_COPYTOFILE;
        if (IsDlgButtonChecked(hDlg, IDC_VERIFYDISK) == BST_CHECKED)
          lpDiskOps->method = IDC_VERIFYDISK;
        if (IsDlgButtonChecked(hDlg, IDC_ERASEDISK) == BST_CHECKED)
          lpDiskOps->method = IDC_ERASEDISK;
        if (IsDlgButtonChecked(hDlg, IDC_CLEARSPACE) == BST_CHECKED)
          lpDiskOps->clear = true;
        else
          lpDiskOps->clear = false;

        // Fill in the dialog box parameter
        lpDiskOps->hDlg = hDlg;
        
        // If a previous thread exists, then close it
        if (hThread)
          CloseHandle(hThread);
        
        // Run the thread
        stopDiskOps = false;
        hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) DiskOps, (LPVOID) lpDiskOps, 0, &threadID);
        if (!hThread)
          DisplayError(hDlg, 1, "Thread creation error", "The selected operation will not run.", true);
      }
      break;
    }

    case IDCANCEL :
      if (threadID) {
        // "Push" the Stop button
        SendMessage(hDlg, WM_COMMAND, MAKEWPARAM(IDC_RUN, BN_PUSHED), NULL);
      } else
        // End the dialog box
        EndDialog(hDlg, 0);
      break;

    case IDC_BROWSEFILES : {
      // Get the filename
      LPOPENFILENAME lpOfn = (LPOPENFILENAME) GlobalAlloc(GPTR, sizeof(OPENFILENAME));
      lpOfn->lStructSize = sizeof(OPENFILENAME);
      lpOfn->hwndOwner = hDlg;
      lpOfn->lpstrFilter = "Disk Images (*.disk, *.dsk, *.txt)\0*.disk;*.dsk;*.txt\0All Files (*.*)\0*.*\0";
      lpOfn->lpstrFile = (LPTSTR) GlobalAlloc(GPTR, sizeof(char) * SendDlgItemMessage(hDlg, IDC_IMAGEPATH, EM_LINELENGTH, 0, 0) + 1024);
      GetDlgItemText(hDlg, IDC_IMAGEPATH, lpOfn->lpstrFile, GlobalSize(GlobalHandle(lpOfn->lpstrFile)));
      lpOfn->nMaxFile = GlobalSize(GlobalHandle(lpOfn->lpstrFile));
      lpOfn->Flags = OFN_PATHMUSTEXIST | OFN_HIDEREADONLY;
      lpOfn->lpstrDefExt = "Disk Image";
      lpOfn->lpstrTitle = "Select File";

      // If the user put in a file, then update the image path
      if (GetOpenFileName(lpOfn))
        SetDlgItemText(hDlg, IDC_IMAGEPATH, lpOfn->lpstrFile);

      // Free the string and the structure
      GlobalFree(GlobalHandle(lpOfn->lpstrFile));
      GlobalFree(GlobalHandle(lpOfn));
      break;
    }

    case IDC_ABOUT :
      // Display the About message
      MessageBox(hDlg, "Disk Imager 1.00.98.2 for Windows NT\n\n(c) 1998 Steven Lawrance\n\nhttp://nt4sg1.dyn.ml.org\nsteve@nt4sg1.dyn.ml.org", "About", MB_OK | MB_ICONASTERISK);
      break;
      
/*  case IDC_DISKDRIVE :
      if (HIWORD(wParam) == CBN_SELCHANGE) {
        // Get the disk path
        LPTSTR lpszDiskPath;
        HANDLE hDiskFile;
        lpszDiskPath = (LPTSTR) GlobalAlloc(GPTR, sizeof(char) * 7); // Example: "\\\\.\\a:", 7 chars w/ terminating NULL
        SendDlgItemMessage(hDlg, IDC_DISKDRIVE, CB_GETLBTEXT, SendDlgItemMessage(hDlg, IDC_DISKDRIVE, CB_GETCURSEL, 0, 0), (LPARAM) lpszDiskPath);
        lpszDiskPath[4] = lpszDiskPath[2];
        CopyMemory(lpszDiskPath, "\\\\.\\", 4);
        lpszDiskPath[5] = ':';
        lpszDiskPath[6] = 0;
        
        // Open the disk for reading
        hDiskFile = CreateFile(lpszDiskPath, GENERIC_READ, FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);
        if (hDiskFile != INVALID_HANDLE_VALUE) {
          // Get the disk geometry
          DISK_GEOMETRY geometry;
          DWORD writtenBytes;
          if (DeviceIoControl(hDiskFile, IOCTL_DISK_GET_DRIVE_GEOMETRY, NULL, 0, &geometry, sizeof(geometry), &writtenBytes, NULL)) {
            // Format the message for the size and display it
            LPTSTR sizeStr = (LPTSTR) GlobalAlloc(GPTR, 25);
            wsprintf(sizeStr, "Size: %lu byte(s)", geometry.TracksPerCylinder * geometry.SectorsPerTrack * geometry.BytesPerSector * geometry.Cylinders.LowPart);
            SetDlgItemText(hDlg, IDC_DISKSIZE, sizeStr);
            GlobalFree(GlobalHandle(sizeStr));
          } else
            SetDlgItemText(hDlg, IDC_DISKSIZE, "");
          
          // Close the disk
          CloseHandle(hDiskFile);
        } else
          SetDlgItemText(hDlg, IDC_DISKSIZE, "");
        
        // Free the disk path memory
        GlobalFree(GlobalHandle(lpszDiskPath));
      }
      break;*/

    }
    
    return 0;
    
  }

  return 0;
}
 

int FAR PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrevInstance, LPSTR lpszCmdLine, int nCmdShow) {
  // Create and run the dialog box
  InitCommonControls();
  threadID = 0;
  hThread = 0;
  if (DialogBoxParam(hInst, (LPSTR) IDD_DISKOPS, HWND_DESKTOP, (DLGPROC) DiskOpsDialogProc, (LPARAM) lpszCmdLine) == -1)
    MessageBox(HWND_DESKTOP, "Error running dialog box!", "Cannot start program", MB_OK | MB_ICONSTOP);

  // Clean up thread
  if (hThread)
    CloseHandle(hThread);
  return 0;
}
